#include "libs/pwm.h"
#include "libs/usart.h"
#include "libs/spi.h"

// #define DEBUG 1
#define BAUD 9600L

#define STATE_KEYBOARD 0
#define STATE_RF 1
#define STATE_WEB_SERIAL 2
#define STATE_WEB_RF 3
#define DEFAULT_STATE STATE_RF

#define PIN_TAIL_MOTOR PWM_0 //OC0A
#define PIN_CENTER_MOTOR PWM_1 //OC0B
#define PIN_TAIL_SERVO PWM_2 //OC1A

#define MIN_INPUT_TAIL_MOTOR 0
#define MAX_INPUT_TAIL_MOTOR 27
#define MIN_OUTPUT_TAIL_MOTOR 42
#define MAX_OUTPUT_TAIL_MOTOR 70

#define MIN_INPUT_CENTER_MOTOR 0
#define MAX_INPUT_CENTER_MOTOR 32
#define MIN_OUTPUT_CENTER_MOTOR 42
#define MAX_OUTPUT_CENTER_MOTOR 75

#define MIN_INPUT_TAIL_SERVO 0
#define MAX_INPUT_TAIL_SERVO 180
#define MIN_OUTPUT_TAIL_SERVO 750
#define MAX_OUTPUT_TAIL_SERVO 3080

#define CHAR_NEWLINE 10
#define BUFFER_SIZE 128

#define SWITCH_STATES 'c'
#define TYPE_MOTOR 'm'
#define TYPE_SERVO 't'

#define CENTER_MOTOR_FASTER 'w'
#define CENTER_MOTOR_SLOWER 's'
#define CENTER_MOTOR_STOP 'q'
#define TAIL_MOTOR_FASTER 'i'
#define TAIL_MOTOR_SLOWER ','
#define TAIL_MOTOR_STOP 'a'
#define TAIL_SERVO_LEFT 'j'
#define TAIL_SERVO_RIGHT 'l'
#define TAIL_SERVO_CENTER 'k'
#define STOP_ALL_MOTORS 'z'

#define MOTOR_STEP 2
#define SERVO_STEP 10

#define TYPE_INDEX 0
#define NUMBER_INDEX 1
#define VALUE_INDEX 3

#ifndef TRUE
   #define TRUE 1
#endif

#ifndef FALSE
   #define FALSE 0
#endif

void handleChar(char c);

// type[# value]

// examples:
// m1 90 (motor 1, set power to 90)
// s1 90 (servo 1, set angle to 90)
// c (switch states)

int state;
char serialInput[BUFFER_SIZE];
int serialIndex; 
char tmpChar;

int centerMotorPower;
int tailMotorPower;
int tailServoAngle;

char buffer[BUFFER_SIZE];
int status;
int readingSerial;




#define MIN_SERVO_INPUT 0
#define MAX_SERVO_INPUT 180
#define MIN_TAIL_SERVO 750
#define MAX_TAIL_SERVO 3080

void setup() {
   disableClockPrescaling();
   USART_initSerial(F_CPU, BAUD); 
   printf("\nInitializing...\n");

   // initialize PWM pins and value mapping
   initTimer(TIMER_0, UNIT_A_AND_B, TIMER_0_DIVIDE_BY_256, TIMER_0_MODE_PHASE_CORRECT);
   initTimer(TIMER_1, UNIT_A, TIMER_1_DIVIDE_BY_8, TIMER_1_MODE_PHASE_AND_FREQUENCY_CORRECT);

   setPWMValueMapping(PIN_TAIL_MOTOR, MIN_INPUT_TAIL_MOTOR, MAX_INPUT_TAIL_MOTOR, MIN_OUTPUT_TAIL_MOTOR, MAX_OUTPUT_TAIL_MOTOR); 
   setPWMValueMapping(PIN_CENTER_MOTOR, MIN_INPUT_CENTER_MOTOR, MAX_INPUT_CENTER_MOTOR, MIN_OUTPUT_CENTER_MOTOR, MAX_OUTPUT_CENTER_MOTOR);
   setPWMValueMapping(PIN_TAIL_SERVO, MIN_INPUT_TAIL_SERVO, MAX_INPUT_TAIL_SERVO, MIN_OUTPUT_TAIL_SERVO, MAX_OUTPUT_TAIL_SERVO); 
   printf("+ PWM pins initialized\n");

   // initialize SPI in slave mode
   SPI_initSlave();
   printf("+ SPI SLAVE mode initialized\n");

   // set default values
   state = DEFAULT_STATE;
   serialIndex = 0;
   centerMotorPower = MIN_INPUT_CENTER_MOTOR;
   tailMotorPower = MIN_INPUT_TAIL_MOTOR;
   tailServoAngle = 90; // center
   readingSerial = FALSE;

   // remove safety breaks
   setPWMValue(PIN_TAIL_MOTOR, tailMotorPower);
   setPWMValue(PIN_CENTER_MOTOR, centerMotorPower);

   // set servo at the center
   setPWMValue(PIN_TAIL_SERVO, tailServoAngle);
   printf("+ Default values set\n");

   #ifdef DEBUG 
    printf("initialization complete\n\n");
    printf("Protocol:\n");
    printf("type[# value]\n\n");
    printf("examples:\n");
    printf("m1 90 (motor 1, set power to 90)\n");
    printf("s1 90 (servo 1, set angle to 90)\n");
    printf("c (switch states)\n");
   #endif


   printf("Hovercraft initialized! \n");
}


/**
 * Handle incoming commands
 *
 * @param *input
 * @param size The length of input
 */
void handleInputCommand(char *input, int size) {
   if (size == 0) {
      return;
   }
  
   char type = input[TYPE_INDEX];
   #ifdef DEBUG
      printf("Type: %c\n", type);
   #endif
  
   if (type == SWITCH_STATES) {
      // change states
      state = STATE_RF;
      #ifdef DEBUG
         printf("-- switching to RF state\n");
      #endif
   } else if (size > VALUE_INDEX) {
    int protocolPass = FALSE;
    int outputDevice;
    int value = atoi(input + VALUE_INDEX);

    // select output device
    if (type == TYPE_MOTOR) {
      protocolPass = TRUE;
      if (input[NUMBER_INDEX] == '1') {
        outputDevice = PIN_TAIL_MOTOR;
        tailMotorPower = value;
        #ifdef DEBUG
          printf("-- tail motor selected\n");
        #endif
      } else {
        outputDevice = PIN_CENTER_MOTOR;
        centerMotorPower = value;
        #ifdef DEBUG
          printf("-- center motor selected\n");
        #endif
      }
    } else if (type == TYPE_SERVO) {
      protocolPass = TRUE;
      outputDevice = PIN_TAIL_SERVO;
      tailServoAngle = value;
      #ifdef DEBUG
        printf("-- tail servo selected\n");
      #endif
    }
    
    if (protocolPass) {
      // set output device value
      #ifdef DEBUG
        printf("value: %d\n", value);
      #endif
      
      setPWMValue(outputDevice, value);
    } else {
      #ifdef DEBUG 
         printf("bad input format\n");
      #endif
    }
  }  
}

/**
 * Parse serial input and pass it forward to the handling module
 **/
void handleKeyboard() {  
  if (USART_isReceiveReady()) {
      USART_receive(&tmpChar);      
      handleChar(tmpChar);
  }
}

/**
 * Parse radio input and pass it forward to the handling module
 **/
void handleRF() {  
 if (SPI_isReceiveReady()) {
      status = SPI_receive(buffer, BUFFER_SIZE);
      
      // be forgiving, we're broadcasting the same character multiple times
      // so chances are we got one right
      #ifdef DEBUG 
          printf("status: %d, received: %s\n", status, buffer);
      #endif
        
      handleChar(buffer[0]);
      
//      if (status == SUCCESS) {
//          if (DEBUG) {
//            Serial.print("received: ");
//            Serial.println(buffer);  
//          }
//          
//          handleChar(buffer[0]);
//      } else {
//          Serial.println("transmission failure!");
//      }
  }
}

///**
// * Parse radio input and pass it forward to the handling module
// **/
//void handleWebRF() {  
// if (SPI_isReceiveReady()) {
//      status = SPI_receive(buffer, BUFFER_SIZE);
//      if (status == SUCCESS) {
//        if (DEBUG) {
//          Serial.print("received: ");
//          Serial.println(buffer);
//        }
//        
//        handleInputCommand(buffer, BUFFER_SIZE);
//      } else if (DEBUG) {
//        Serial.println("transmission failure!");
//      }
//
//  }
//}


void handleDefaultKey(char c) {  
  serialInput[serialIndex++] = c;
  serialInput[serialIndex] = '\0'; // null terminate the string
  
  if (serialIndex > BUFFER_SIZE - 2) {
    #ifdef DEBUG
      if (USART_isReceiveReady()) {
        printf("exceeded buffer size, ignoring some input!\n");
      }  
    #endif
    
    // clear buffer and saved input
   char tmp;
   while (1) {
      USART_receive(&tmp);
      if (tmp == -1) {
         break;
      }
   }
   
   serialIndex = 0;
   serialInput[serialIndex] = '\0';  
  }
}

void handleChar(char c) {
  switch (c) {
        case CHAR_NEWLINE:
          readingSerial = FALSE;
          handleInputCommand(serialInput, serialIndex);
         
          // clear saved input
          serialIndex = 0;
          serialInput[serialIndex] = '\0';
          break;
        
        case CENTER_MOTOR_FASTER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (centerMotorPower < MAX_INPUT_CENTER_MOTOR) {        
            #ifdef DEBUG 
              printf("- center motor faster\n");
            #endif
            
            centerMotorPower += MOTOR_STEP;
            setPWMValue(PIN_CENTER_MOTOR, centerMotorPower);
          }
          
          break;
          
        case CENTER_MOTOR_SLOWER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          if (centerMotorPower > MIN_INPUT_CENTER_MOTOR) {        
            #ifdef DEBUG
              printf("- center motor slower\n");
            #endif
            
            centerMotorPower -= MOTOR_STEP;
            setPWMValue(PIN_CENTER_MOTOR, centerMotorPower);
          }
          
          break;
          
        case CENTER_MOTOR_STOP:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          #ifdef DEBUG
            printf("- center motor stop\n");
          #endif
          
          centerMotorPower = MIN_INPUT_CENTER_MOTOR;
          setPWMValue(PIN_CENTER_MOTOR, centerMotorPower);
          break;
        
        case TAIL_MOTOR_FASTER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailMotorPower < MAX_INPUT_TAIL_MOTOR) {
            #ifdef DEBUG
              printf("- tail motor faster\n");
            #endif
            
            tailMotorPower += MOTOR_STEP;
            setPWMValue(PIN_TAIL_MOTOR, tailMotorPower);
          }
          
          break;
          
        case TAIL_MOTOR_SLOWER:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailMotorPower > MIN_INPUT_TAIL_MOTOR) {    
            #ifdef DEBUG
              printf("- tail motor slower\n");
            #endif
            
            tailMotorPower -= MOTOR_STEP;
            setPWMValue(PIN_TAIL_MOTOR, tailMotorPower);
          }
          
          break;

        case TAIL_MOTOR_STOP:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          #ifdef DEBUG
            printf("- tail motor stop\n");
          #endif
          
          tailMotorPower = MIN_INPUT_TAIL_MOTOR;
          setPWMValue(PIN_TAIL_MOTOR, tailMotorPower);
          break;
          
        case TAIL_SERVO_LEFT:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          if (tailServoAngle > MIN_INPUT_TAIL_SERVO) {        
            #ifdef DEBUG
              printf("- tail servo left\n");
            #endif
            
            tailServoAngle -= SERVO_STEP;
            setPWMValue(PIN_TAIL_SERVO, tailServoAngle);
          }
          
          break;
        case TAIL_SERVO_RIGHT:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
            
          if (tailServoAngle < MAX_INPUT_TAIL_SERVO) {        
            #ifdef DEBUG
              printf("- tail servo right\n");
            #endif
            
            tailServoAngle += SERVO_STEP;
            setPWMValue(PIN_TAIL_SERVO, tailServoAngle);
          }
          
          break;
          
        case TAIL_SERVO_CENTER:
            if (readingSerial) {
              handleDefaultKey(c);
              break;
            }
            
            #ifdef DEBUG
              printf("- tail servo center\n");
            #endif
            
            tailServoAngle = 90;
            setPWMValue(PIN_TAIL_SERVO, tailServoAngle);
          break;
          
        case STOP_ALL_MOTORS:
          if (readingSerial) {
            handleDefaultKey(c);
            break;
          }
          
          #ifdef DEBUG
            printf("** STOPPING ALL MOTORS **\n");
          #endif
          
          centerMotorPower = MIN_INPUT_CENTER_MOTOR;
          setPWMValue(PIN_CENTER_MOTOR, centerMotorPower);
          tailMotorPower = MIN_INPUT_TAIL_MOTOR;
          setPWMValue(PIN_TAIL_MOTOR, tailMotorPower);
          break;
          
        default:
            if (state != STATE_RF) {
              readingSerial = TRUE;
              handleDefaultKey(c);
            }
      }  
}


void inline loop() {
   switch (state) {
    case STATE_KEYBOARD:
    case STATE_WEB_SERIAL:
      handleKeyboard();
      break;
      
//    case STATE_WEB_RF:
//      handleWebRF();
//      break;

    case STATE_WEB_RF:
    case STATE_RF:
      handleRF();
      break;
  }
}

int main(void) {
   setup();

   while (1) {
      loop();
   }
}