//***************************************************************************
// Hello SPI
// * Make sure the pins are correctly configured for your board
//
// ** assuming there are no reads and writes simultaneously or two types
// ** (polling and interrupt based) are not operating at the time
//
//
// @author Amir Lazarovich 
// @version 0.1 2013
//
// * based on atmel example code
//**************************************************************************

#ifndef __spi_v1__
#define __spi_v1__

#if defined(ARDUINO) && ARDUINO >= 100
        #include <Arduino.h>
#elif defined(ARDUINO)
        #include <WProgram.h>
#else
        #include "avr-io.h"
#endif

#include <avr/interrupt.h>
#include <string.h>
#include <ctype.h>

#define ERROR_TRANSMISSION_FAILED -2
#define ERROR_BUFFER_SIZE_TOO_SMALL -1
#define SUCCESS 1

#define NUM_OF_ATTEMPTS 10
#define ERROR_CHAR '$'
#define END_TRANSMISSION_CHAR '@'

int attempts;
char* __pbuffer;
unsigned char __size;

void SPI_initMaster(void);
void SPI_initSlave(void);
void SPI_initSlaveIntr(void);
inline void SPI_sendChar(char c);
void SPI_send(char* msg);
inline void SPI_sendWithSize(char* msg, unsigned char msgSize);
void SPI_sendIntr(char* msg);
inline void SPI_sendIntrWithSize(char* msg, unsigned char msgSize);
inline int SPI_isReceiveReady();
inline char SPI_receiveChar();
inline int SPI_receive(char* buffer, unsigned char bufferSize); 
inline void SPI_handleReceiveInterrupt(char* buffer, unsigned char bufferSize, void (*callback)(char*, int));
inline void SPI_handleSendInterrupt(void (*callback)(void));
// *********************** inline usart.c ***********************

/**
 * Intialization Routine Master Mode (polling)
 */
void SPI_initMaster(void) {
        volatile char IOReg;
        // set PB2(/SS), PB3(MOSI), PB5(SCK) as output 
        DDRB |= (1<<PINB2)|(1<<PINB3)|(1<<PINB5);

        // enable SPI in Master Mode with SCK = CK/4
        SPCR  = (1<<SPE)|(1<<MSTR);
        // SPCR = (1<<SPE)|(1<<MSTR)|(0<<CPHA)|(0<<CPOL)|(0<<SPR1)|(0<<SPR0);

        // clear SPIF bit in SPSR
        IOReg = SPSR;                 
        IOReg = SPDR;
}

void SPI_initMasterIntr(void) {
        SPI_initMaster();
        
        // enable SPI Interrupt and set SCK = CK/16
        SPCR |= (1<<SPIE)|(1<<SPR0);
        
        // enable global interrupts
        sei();
}

/** 
 * Intialization Routine Slave Mode (polling)
 */
void SPI_initSlave(void) {
        volatile char IOReg;
        // set PB4(MISO) as output 
        DDRB |= (1<<PINB4);

        // enable SPI in Slave Mode with SCK = CK/4
        SPCR = (1<<SPE);
        // SPCR = (1<<SPE)|(0<<CPHA)|(0<<CPOL)|(0<<SPR1)|(0<<SPR0);

        // clear SPIF bit in SPSR
        IOReg = SPSR;                 
        IOReg = SPDR;
}

void SPI_disableSlaveIntr(void) {
        // disable SPI Interrupt
        SPCR &= (0<<SPIE);
        
        // disable global interrupts
        cli();
}


void SPI_enableSlaveIntr(void) {
        // enable SPI Interrupt
        SPCR |= (1<<SPIE);
        
        // enable global interrupts
        sei();
}

void SPI_initSlaveIntr(void) {
        SPI_initSlave();
        SPI_enableSlaveIntr();
}

inline void SPI_sendChar(char c) {
        // send character
        SPDR = c;

        // wait until character is sent
        while (!(SPSR & (1<<SPIF))); 
}

void SPI_send(char* msg) {
        SPI_sendWithSize(msg, strlen(msg));
}

inline void SPI_sendWithSize(char* msg, unsigned char msgSize) {
        if (msgSize > 0) {
                SPI_sendChar(msgSize);
                
                __pbuffer = msg;             
                while (msgSize-- > 0) {    
                        SPI_sendChar(*__pbuffer);

                        // point to next char in msg
                        __pbuffer++;
                }

                SPI_sendChar(END_TRANSMISSION_CHAR);
        }
}

void SPI_sendIntr(char* msg) {
        SPI_sendIntrWithSize(msg, strlen(msg));
}

inline void SPI_sendIntrWithSize(char* msg, unsigned char msgSize) {
        if (msgSize > 0) {
                __pbuffer = msg;
                __size = msgSize;
                SPDR = msgSize; 
        }
}

inline int SPI_isReceiveReady() {
        return (SPSR & (1<<SPIF)) != 0; 
}

// TODO use a timer so won't get stuck on an infinite loop
inline char SPI_receiveChar() {
        // Wait until character is received
        attempts = NUM_OF_ATTEMPTS;
        while (!(SPSR & (1<<SPIF)) && attempts-- > 0);
        return SPDR;
        // if (attempts < 0) {
        //     return SPDR;
        //     // clean register
        //     // volatile char IOReg;
        //     // IOReg = SPDR;

        //     // return error character
        //     // return ERROR_CHAR;
        // } else {
        //     return SPDR;     
        // }      
}

#define STATE_IGNORE_SIZE 1
#define STATE_NORMAL 0

inline int SPI_receive(char* buffer, unsigned char bufferSize) {
        if (bufferSize > 1) {
                // get message size
                __size = SPI_receiveChar();
                // if (attempts <= 0) {
                //     return ERROR_TRANSMISSION_FAILED;
                // }

                // check that we received the message size correctly
                // if not, lets use the buffer size. Otherwise,
                // check that given buffer is large enough
                __pbuffer = buffer;
                unsigned short state = STATE_NORMAL;
                if (isalpha(__size)) {
                    // lets assume the first character belongs to the message
                    // it may fail sometimes
                    *__pbuffer = __size;
                    __pbuffer++;

                    __size = bufferSize;
                    state = STATE_IGNORE_SIZE;
                }
                else if (__size > bufferSize - 1) {
                        return ERROR_BUFFER_SIZE_TOO_SMALL; 
                }

                // start reading the message
                while (__size-- > 0) {
                        *__pbuffer = SPI_receiveChar();
                        if (*__pbuffer == END_TRANSMISSION_CHAR) {
                            if (state == STATE_NORMAL) {
                                return ERROR_TRANSMISSION_FAILED;    
                            } else {
                                break;
                            }
                        }

                        // if (attempts <= 0) {
                        //     return ERROR_TRANSMISSION_FAILED;
                        // }

                        __pbuffer++;
                }

                if (state == STATE_NORMAL) {
                    // read last sufix
                    SPI_receiveChar();
                }
        } else {
            return ERROR_BUFFER_SIZE_TOO_SMALL; 
        }

        *__pbuffer = '\0';
        return SUCCESS;
}

    
/**
 * Handle received transmission interrupt event
 * 
 * @param buffer where we'll save the transmission
 * @param bufferSize our maxmium transmission size - if we detect a message that has a larger value we will ignore it and return a "failure" status
 * @param callback callback that will be invoked when transission is over
 */
inline void SPI_handleReceiveInterrupt(char* buffer, unsigned char bufferSize, void (*callback)(char*, int)) {
        if (__size == 0) {
            __size = SPDR;

            // check that given buffer is large enough
            if (__size > bufferSize - 1) {
                    (*callback)(NULL, ERROR_BUFFER_SIZE_TOO_SMALL);       
            } else {
                    __pbuffer = buffer;        
            }
        } else {
            // append one more byte
            *__pbuffer = SPDR;
            __pbuffer++;
            *__pbuffer = '\0';
            __size--;    
            
            if (__size == 0) {
                    // that's it, notify we're done
                    (*callback)(buffer, SUCCESS);       
            }       
        }
}

inline void SPI_handleSendInterrupt(void (*callback)(void)) {
        if (__size > 0) {
                // send character
                SPDR = *__pbuffer; 

                __pbuffer++;
                __size--;
        } else {
                // finished
                (*callback)();
        }
}

#endif